'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _copy = require('./utils/copy');

var _copy2 = _interopRequireDefault(_copy);

var _$traverse = require('./$traverse');

var _$traverse2 = _interopRequireDefault(_$traverse);

var _reduceSequence = require('./utils/reduceSequence');

var _reduceSequence2 = _interopRequireDefault(_reduceSequence);

var _$none = require('./$none');

var _removed = require('./utils/removed');

var _removed2 = _interopRequireDefault(_removed);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var $each = (0, _$traverse2.default)({
  select: function select(object, next) {
    // Pass each value along to the next navigator.
    return (0, _reduceSequence2.default)(function (result, key) {
      return next(object[key]);
    }, undefined, object);
  },
  update: function update(object, next) {
    var isArray = Array.isArray(object);
    // Marker for if we removed values from an array. Since we filter out
    // removed values, we don't want to do that iteration if we don't have to.
    // Maybe there's a better way to do this, but removing array values in place
    // is kind of ugly, since positions change.
    var didRemoveValues = false;
    var newObject = (0, _reduceSequence2.default)(function (result, key) {
      var newValue = next(object[key]);
      var isNewValueNone = (0, _$none.isNone)(newValue);
      // This is a no-op unless we actually have a different value. The
      // `isNoneValue` covers a goofy edge where you might have the value $none
      // stored in the object/array. In that case, it would match, but we still
      // want to remove it.
      if (newValue !== object[key] || isNewValueNone) {
        // Create a new object if we haven't done that yet.
        if (object === result) {
          result = (0, _copy2.default)(result);
        }

        if (isNewValueNone) {
          // For arrays, we'll store a removed value so indexes don't get wonky.
          if (isArray) {
            result[key] = _removed2.default;
            didRemoveValues = true;
            // Otherwise, we'll just delete it now.
          } else {
            delete result[key];
          }
        } else {
          result[key] = newValue;
        }
      }
      return result;
    }, object, object);
    // Filter our removed values from arrays.
    if (isArray && didRemoveValues) {
      return newObject.filter(_removed.isNotRemoved);
    }
    return newObject;
  }
});

exports.default = $each;