'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _copy = require('./utils/copy');

var _copy2 = _interopRequireDefault(_copy);

var _$traverse = require('./$traverse');

var _$traverse2 = _interopRequireDefault(_$traverse);

var _reduceSequence = require('./utils/reduceSequence');

var _reduceSequence2 = _interopRequireDefault(_reduceSequence);

var _$none = require('./$none');

var _removed = require('./utils/removed');

var _removed2 = _interopRequireDefault(_removed);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var $eachKey = (0, _$traverse2.default)({
  select: function select(object, next) {
    // Pass each key along to the next navigator.
    return (0, _reduceSequence2.default)(function (result, key) {
      return next(key);
    }, undefined, object);
  },
  update: function update(object, next) {
    var isArray = Array.isArray(object);
    // Marker for if we removed keys from an array. Since we filter out
    // removed keys, we don't want to do that iteration if we don't have to.
    // Maybe there's a better way to do this, but removing array values in place
    // is kind of ugly, since positions change.
    var didRemoveKeys = false;
    var newObject = (0, _reduceSequence2.default)(function (result, key) {
      var newKey = next(key);
      // Undefined/null key doesn't make much sense, so treat it as $none.
      var isKeyNone = (0, _$none.isNone)(newKey) || newKey == null;
      // Marker for if we removed values from an array. Since we filter out
      // removed values, we don't want to do that iteration if we don't have to.
      // Maybe there's a better way to do this, but removing array values in place
      // is kind of ugly, since positions change.
      didRemoveKeys = didRemoveKeys || isKeyNone;
      // This is a no-op unless we actually have a different key.
      if (newKey !== key) {
        // Create a new object if we haven't done that yet.
        if (object === result) {
          result = (0, _copy2.default)(result);
        }
        // For arrays, we'll store a removed value so indexes don't get wonky.
        if (isArray) {
          result[key] = _removed2.default;
          didRemoveKeys = true;
          // Otherwise, we'll just delete it now.
        } else {
          delete result[key];
        }
        // If we have a new key, point the new key to the value of the old key.
        if (!isKeyNone) {
          result[newKey] = object[key];
        }
      }
      return result;
    }, object, object);
    // Filter our removed values from arrays.
    if (isArray && didRemoveKeys) {
      return newObject.filter(_removed.isNotRemoved);
    }
    return newObject;
  }
});

exports.default = $eachKey;