'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _copy = require('./utils/copy');

var _copy2 = _interopRequireDefault(_copy);

var _$traverse = require('./$traverse');

var _$traverse2 = _interopRequireDefault(_$traverse);

var _reduceSequence = require('./utils/reduceSequence');

var _reduceSequence2 = _interopRequireDefault(_reduceSequence);

var _$none = require('./$none');

var _removed = require('./utils/removed');

var _removed2 = _interopRequireDefault(_removed);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var $eachPair = (0, _$traverse2.default)({
  select: function select(object, next) {
    // Pass each pair along to the next navigator.
    return (0, _reduceSequence2.default)(function (result, key) {
      return next([key, object[key]]);
    }, undefined, object);
  },
  update: function update(object, next) {
    var isArray = Array.isArray(object);
    // Marker for if we removed pairs from an array. Since we filter out
    // removed pairs, we don't want to do that iteration if we don't have to.
    // Maybe there's a better way to do this, but removing array values in place
    // is kind of ugly, since positions change.
    var didRemovePairs = false;
    var newObject = (0, _reduceSequence2.default)(function (result, key) {
      var value = object[key];
      var pair = [key, value];
      var newPair = next(pair);
      var newKey = void 0;
      var newValue = void 0;
      // Undefined/null pair will result in undefined key, which will be
      // treated as $none.
      if (newPair != null) {
        newKey = newPair[0];
        newValue = newPair[1];
      }
      var isPairNone =
      // Undefined/null key doesn't make much sense, so treat it as $none.
      newKey == null || (0, _$none.isNone)(newPair) ||
      // isNone(newValue) covers a goofy edge where you might have the value
      // $none stored in an object/array. In that case, it would match, but we
      // still want to remove it.
      (0, _$none.isNone)(newValue);
      // This is a no-op unless we actually have a different value or key.
      if (newKey !== key || newValue !== value || isPairNone) {
        // Create a new object if we haven't done that yet.
        if (object === result) {
          result = (0, _copy2.default)(result);
        }
        if (newKey !== key) {
          // For arrays, we'll store a removed value so indexes don't get wonky.
          if (isArray) {
            result[key] = _removed2.default;
            didRemovePairs = true;
            // Otherwise, we'll just delete it now.
          } else {
            delete result[key];
          }
        }
        // If we haven't removed the key or value, then set the new key to the
        // new value.
        if (!isPairNone) {
          result[newKey] = newValue;
        }
      }
      return result;
    }, object, object);
    // Filter our removed values from arrays.
    if (isArray && didRemovePairs) {
      return newObject.filter(_removed.isNotRemoved);
    }
    return newObject;
  }
});

exports.default = $eachPair;