'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.traverseEach = undefined;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _objectAssign = require('object-assign');

var _objectAssign2 = _interopRequireDefault(_objectAssign);

var _copy = require('./utils/copy');

var _copy2 = _interopRequireDefault(_copy);

var _$traverse = require('./$traverse');

var _$nav = require('./$nav');

var _arrayify = require('./utils/arrayify');

var _arrayify2 = _interopRequireDefault(_arrayify);

var _$set = require('./$set');

var _$default = require('./$default');

var _$apply = require('./$apply');

var _$lens = require('./$lens');

var _$setContext = require('./$setContext');

var _$none = require('./$none');

var _$none2 = _interopRequireDefault(_$none);

var _reduced = require('./utils/reduced');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// `traverseEach` is the heart of `qim`. It's a little ugly because it needs to
// be performant, and it's generic across select and update.
var traverseEach = exports.traverseEach = function traverseEach(navKey, // select or update?
state, // state to carry around, so select can add new values
resultFn, // like a reduce function, called for each selected value
path, // the whole query
object, // the current object that we're traversing
pathIndex, // index into the path
returnFn, // function to "return" to, if we've navigated into a sub-query
context, // current context: modified with setContext and passed to transforms
mutationMarker // performance signal that it's safe to mutate instead of clone
) {

  if (pathIndex >= path.length) {
    // If we're in a sub-query (with $nav), we need to return.
    if (returnFn) {
      return returnFn(object, context);
    }
    // Default to identity if there's no result function.
    return resultFn ? resultFn(state, object) : object;
  }

  var nav = path[pathIndex];

  // null/undefined selects/updates nothing.
  if (nav == null) {
    return navKey === _$traverse.selectKey ? _$none2.default : object;
  }

  // Primitives are handled right here inside traverseEach to stay efficient.
  if (typeof nav === 'string' || typeof nav === 'number' || typeof nav === 'boolean') {
    // select, which is pretty simple
    if (navKey === _$traverse.selectKey) {
      // If we have an object, just grab the nested object and keep digging.
      if (object && (typeof object === 'undefined' ? 'undefined' : _typeof(object)) === 'object') {
        var subObject = object[nav];
        // Special case so `has` can differentiate between missing keys and keys
        // pointing to undefined values. Maybe a better way? Maybe just don't
        // worry about `has(['x'], {x: undefined}) === false`?
        if (typeof subObject === 'undefined' && !(nav in object) && pathIndex === path.length - 1) {
          return _$none2.default;
        }
        return traverseEach(navKey, state, resultFn, path, subObject, pathIndex + 1, returnFn, context);
      } else {
        return _$none2.default;
      }
    }
    // update, which is not as simple
    if (object && (typeof object === 'undefined' ? 'undefined' : _typeof(object)) === 'object') {
      var value = object[nav];
      // Dig in for the new value.
      var newValue = traverseEach(navKey, state, resultFn, path, value, pathIndex + 1, returnFn, context);
      // A new value of $none means we've removed it.
      if ((0, _$none.isNone)(newValue)) {
        if (!(nav in object)) {
          return object;
        }
        if (Array.isArray(object)) {
          var _newObject2 = object.slice(0);
          // Could have a silly edge here where nav was a property and not an
          // index, which means it disappears. If it's still here, that means
          // it's an index.
          if (nav in object) {
            _newObject2.splice(nav, 1);
          }
          return _newObject2;
        }
        var _newObject = (0, _objectAssign2.default)({}, object);
        delete _newObject[nav];
        return _newObject;
      }
      // If we got back the same value, it's a no-op.
      if (value === newValue) {
        return object;
      }
      // Some dangerous mutation going on here! Be careful!
      if (mutationMarker) {
        // If we have a mutation marker, and we've already mutated, then we can
        // safely mutate and avoid the cost of re-cloning the object.
        if (mutationMarker.hasMutated) {
          object[nav] = newValue;
          return object;
          // Otherwise, we're going to mutate, so mark that we've mutated.
        } else {
          mutationMarker.hasMutated = true;
        }
      }
      // Clone our object and set the property to the new value.
      var newObject = (0, _copy2.default)(object);
      newObject[nav] = newValue;
      return newObject;
      // If we got back null/undefined, then intelligently create a new object.
      // It's debatable on whether or not this is a good idea. Integers used to
      // auto-create arrays, but that's probably wrong more than right. An
      // integer like 123456 is probably an id and not an array index, so it's
      // probably a mistake to create an array of that size. If an array is
      // actually wanted, `default([])` can be added. We could do nothing, which
      // would keep you from making typos, but then you have to add a lot of
      // `default({})`, and then you're right back to being able to make typos.
      // So defaulting to create an object seems like a sane balance.
    } else if (object == null) {
      var _newValue = traverseEach(navKey, state, resultFn, path, undefined, pathIndex + 1, returnFn, context);
      // Trying to set a key to $none is a no-op.
      if ((0, _$none.isNone)(_newValue)) {
        return object;
      }
      var _newObject3 = {};
      _newObject3[nav] = _newValue;
      return _newObject3;
    } else {
      throw new Error('Cannot update property ' + nav + ' (at path index ' + pathIndex + ') for non-object.');
    }
  }

  // Predicate navigator, keep going if there's a match.
  if (typeof nav === 'function') {
    if (nav(object)) {
      return traverseEach(navKey, state, resultFn, path, object, pathIndex + 1, returnFn, context);
    } else {
      return navKey === _$traverse.selectKey ? _$none2.default : object;
    }
  }

  // Some built-in navigators for transform and context.
  switch (nav['@@qim/nav']) {
    // Transform the current value into another value.
    case _$apply.$applyKey:
      {
        return traverseEach(navKey, state, resultFn, path, nav.fn(object, context), pathIndex + 1, returnFn, context);
      }
    // Transform the current value into another value.
    // For updates, apply another transform on the way back.
    case _$lens.$lensKey:
      {
        var result = traverseEach(navKey, state, resultFn, path, nav.fn(object, context), pathIndex + 1, returnFn, context);
        if (navKey === _$traverse.selectKey || !nav.fromFn) {
          return result;
        }
        return nav.fromFn(result, object, context);
      }
    // Set the current value to a new constant value.
    case _$set.$setKey:
      return traverseEach(navKey, state, resultFn, path, nav.value, pathIndex + 1, returnFn, context);
    // Set the current value to a new constant value if the current value is
    // undefined.
    case _$default.$defaultKey:
      {
        if (typeof object === 'undefined') {
          return traverseEach(navKey, state, resultFn, path, nav.value, pathIndex + 1, returnFn, context);
        }
        return traverseEach(navKey, state, resultFn, path, object, pathIndex + 1, returnFn, context);
      }
    // Remove the current value.
    case _$none.$noneKey:
      return _$none2.default;
    // Set a context value, to later be used by $apply.
    case _$setContext.$setContextKey:
      {
        context = nav.setContext(nav, nav.fn(object, context || {}), context);
        return traverseEach(navKey, state, resultFn, path, object, pathIndex + 1, returnFn, context);
      }
  }

  var navPath = nav[_$nav.pathKey];

  // TODO: Combine this with array navigator.
  // Path navigators.
  if (navPath) {
    // A path navigator can potentially go down multiple paths.
    var moreNavPaths = nav.moreNavPaths;
    var moreNavPathsIndex = 0;
    var navResult = object;
    do {
      // If it's a function, get our path dynamically.
      if (typeof navPath === 'function') {
        navPath = nav.hasArgs ? navPath(nav.args, object, context) : navPath(object, context);
        navPath = (0, _arrayify2.default)(navPath);
      }
      // TODO: Use mutationMarker.
      // An empty query just means continue.
      if (navPath.length === 0) {
        navResult = traverseEach(navKey, state, resultFn, path, object, pathIndex + 1, returnFn, context);
        // Otherwise, recurse in with our new query, but pass a `returnFn` so we can
        // continue where we left off.
      } else {
        navResult = traverseEach(navKey, state, resultFn, navPath, object, 0, function (_object, _context) {
          return traverseEach(navKey, state, resultFn, path, _object, pathIndex + 1, returnFn, _context);
        }, context);
      }
      if (moreNavPaths) {
        // Selects should be applied to the same object, since nothing has changed. If we've reduced, then we're done.
        if (navKey === _$traverse.selectKey) {
          if ((0, _reduced.isReduced)(navResult)) {
            return navResult;
          }
          // Updates should be applied to the new object.
        } else {
          object = (0, _$none.undefinedIfNone)(navResult);
        }
        navPath = moreNavPaths[moreNavPathsIndex];
        moreNavPathsIndex++;
      } else {
        navPath = undefined;
      }
    } while (navPath);
    return navResult;
  }

  // Core select/update navigator.
  if (nav[navKey]) {
    if (nav.hasArgs) {
      return nav[navKey](nav.args, object, function (subObject) {
        return traverseEach(navKey, state, resultFn, path, subObject, pathIndex + 1, returnFn, context);
      }, path, pathIndex);
    }
    return nav[navKey](object, function (subObject) {
      return traverseEach(navKey, state, resultFn, path, subObject, pathIndex + 1, returnFn, context);
    }, path, pathIndex);
  }

  // Nested sub-query.
  if (Array.isArray(nav)) {
    // A nested select will branch off an add items to the current select state.
    // So it's semantically a bit different than a nested update sub-query.
    if (navKey === _$traverse.selectKey) {
      var subResult = traverseEach(navKey, state, resultFn, nav, object, 0, returnFn, context);
      if (pathIndex + 1 === path.length) {
        return subResult;
      }
      return traverseEach(navKey, state, resultFn, path, object, pathIndex + 1, returnFn, context);
    }

    // A nested update will branch off and mutate the object and then return to
    // the current object. It's kind of just a fancy $apply.
    // Create a mutation marker so we can efficiently apply many sub-queries to
    // the current object.
    mutationMarker = mutationMarker || {
      hasMutated: false
    };
    var nestedResult = (0, _$none.undefinedIfNone)(traverseEach(navKey, state, resultFn, nav, object, 0, undefined, context, mutationMarker));
    return traverseEach(navKey, state, resultFn, path, nestedResult, pathIndex + 1, returnFn, context, mutationMarker);
  }

  throw new Error('Invalid navigator ' + nav + ' at path index ' + pathIndex + '.');
};